/*
 NINJAM AU client
 Copyright (C) 2006 Expert Sleepers
 http://www.expertsleepers.co.uk/
 
 Original NINJAM code Copyright (C) 2005 Cockos Incorporated
 http://www.ninjam.com/
 
 This is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This software is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this software; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ninjamplugin.h"
#include "njinterface.h"

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

COMPONENT_ENTRY(ninjamplugin)


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	ninjamplugin::ninjamplugin
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ninjamplugin::ninjamplugin(AudioUnit component)
	: AUEffectBase(component)
{
	CreateElements();
        
#if AU_DEBUG_DISPATCHER
	mDebugDispatcher = new AUDebugDispatcher (this);
#endif
	
	m_pInterface = new njinterface( NUM_IN_CHANNELS, NUM_OUT_CHANNELS );
	
	UInt32 frames = GetMaxFramesPerSlice();
	for ( int i=0; i<NUM_OUT_CHANNELS; ++i )
		m_pBuffers[i] = new float[ frames ];
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	ninjamplugin::~ninjamplugin
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ninjamplugin::~ninjamplugin()
{ 
#if AU_DEBUG_DISPATCHER
	delete mDebugDispatcher;
#endif
	
	delete m_pInterface;
	
	for ( int i=0; i<NUM_OUT_CHANNELS; ++i )
		delete [] m_pBuffers[i];
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UInt32		ninjamplugin::SupportedNumChannels ( const AUChannelInfo**	outInfo )
{
	static const AUChannelInfo m_aobSupportedNumChannels[] = { { NUM_IN_CHANNELS, NUM_OUT_CHANNELS } };
	
    if ( outInfo != NULL )
        *outInfo = &m_aobSupportedNumChannels[0];
    return ( sizeof m_aobSupportedNumChannels/sizeof(AUChannelInfo) );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	ninjamplugin::GetParameterValueStrings
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ComponentResult		ninjamplugin::GetParameterValueStrings(AudioUnitScope		inScope,
                                                                AudioUnitParameterID	inParameterID,
                                                                CFArrayRef *		outStrings)
{
        
    return kAudioUnitErr_InvalidProperty;
}



//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	ninjamplugin::GetParameterInfo
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ComponentResult		ninjamplugin::GetParameterInfo(AudioUnitScope		inScope,
                                                        AudioUnitParameterID	inParameterID,
                                                        AudioUnitParameterInfo	&outParameterInfo )
{
	ComponentResult result = noErr;

	outParameterInfo.flags = 	kAudioUnitParameterFlag_IsWritable
						|		kAudioUnitParameterFlag_IsReadable;
    
    if (inScope == kAudioUnitScope_Global) {
        switch(inParameterID)
        {
            default:
                result = kAudioUnitErr_InvalidParameter;
                break;
            }
	} else {
        result = kAudioUnitErr_InvalidParameter;
    }
    


	return result;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	ninjamplugin::GetPropertyInfo
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ComponentResult		ninjamplugin::GetPropertyInfo (AudioUnitPropertyID	inID,
                                                        AudioUnitScope		inScope,
                                                        AudioUnitElement	inElement,
                                                        UInt32 &		outDataSize,
                                                        Boolean &		outWritable)
{
	switch ( inID )
	{
		case kPropertyNJInterface:
			outDataSize = sizeof( njinterface* );
			outWritable = false;
			return noErr;
	}
	
	return AUEffectBase::GetPropertyInfo (inID, inScope, inElement, outDataSize, outWritable);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//	ninjamplugin::GetProperty
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
ComponentResult		ninjamplugin::GetProperty(	AudioUnitPropertyID inID,
                                                        AudioUnitScope 		inScope,
                                                        AudioUnitElement 	inElement,
                                                        void *			outData )
{
	switch ( inID )
	{
		case kPropertyNJInterface:
			*(njinterface**)outData = m_pInterface;
			return noErr;
	}

	return AUEffectBase::GetProperty (inID, inScope, inElement, outData);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSStatus			ninjamplugin::ProcessBufferLists(
											   AudioUnitRenderActionFlags &	ioActionFlags,
											   const AudioBufferList &			inBuffer,
											   AudioBufferList &				outBuffer,
											   UInt32							inFramesToProcess )
{
    // only support AU v2 format i.e. non-interleaved
    if ( inBuffer.mNumberBuffers != NUM_IN_CHANNELS || outBuffer.mNumberBuffers != NUM_OUT_CHANNELS )
	{
		dprintf(( "num buffers: %d %d\n", (int)inBuffer.mNumberBuffers, (int)outBuffer.mNumberBuffers ));
        return kAudioUnitErr_FormatNotSupported;
	}

	const float** pIns = (const float**)alloca( NUM_IN_CHANNELS * sizeof(const float*) );

	for ( int i=0; i<NUM_IN_CHANNELS; ++i )
		pIns[i] = (const float*)inBuffer.mBuffers[i].mData;
	
	m_pInterface->AudioProc( pIns, m_pBuffers, NUM_IN_CHANNELS, NUM_OUT_CHANNELS, inFramesToProcess, GetSampleRate() );
	for ( int i=0; i<NUM_OUT_CHANNELS; ++i )
		memcpy( outBuffer.mBuffers[i].mData, m_pBuffers[i], inFramesToProcess*sizeof(float) );

	/*
	while ( inFramesToProcess-- )
	{
		*pOut0++ = *pIn0++;
		*pOut1++ = *pIn1++;
	}
	*/
	
    ioActionFlags &= ~kAudioUnitRenderAction_OutputIsSilence;
	
	return noErr;
}
