/*
 NINJAM AU client
 Copyright (C) 2006 Expert Sleepers
 http://www.expertsleepers.co.uk/
 
 Original NINJAM code Copyright (C) 2005 Cockos Incorporated
 http://www.ninjam.com/
 
 This is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This software is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this software; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include	"njinterface.h"
#include	"njclient.h"
#include	"njmisc.h"

static const char* PREFS_NAME = "com.expertsleepers.ninjam";

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

njinterface::njinterface( int numIns, int numOuts )
: m_runThreadId( 0 )
, m_window( 0 )
, m_ownWindow( false )
, m_sampleRate( 1 )
, m_numIns( numIns )
, m_numOuts( numOuts )
{
	// make client
	m_pClient = new NJClient;
	m_pClient->LicenseAgreementCallback = LicenseCallback;
	m_pClient->LicenseAgreement_User32 = (int)this;
	m_pClient->ChatMessage_Callback = ChatMessageCallback;
	m_pClient->ChatMessage_User32 = (int)this;
	std::string home( getenv( "HOME" ) );
	home += "/Desktop/ninjamplugin/";
    mkdir( home.c_str(), 0700 );
	m_pClient->SetWorkDir( (char*)home.c_str() );
	
	// inputs
	for ( int i=0; i<numIns; ++i )
	{
		float pan = ( i & 1 ) ? 1.0 : -1.0f;
		char name[32];
		sprintf( name, "%d %s", i/2, (i&1) ? "right" : "left" );
		bool broadcast = ( i < 2 );
		m_pClient->SetLocalChannelInfo( i, name, true, i, false, 0, true, broadcast );
		m_pClient->SetLocalChannelMonitoring( i, true, 1.0f, true, pan, true, false, true, false );
	}
	
	m_pClient->config_metronome_channel = numOuts - 1;
	
	// make thread to run client on
	OSStatus err;
	
	if ( !MPLibraryIsLoaded() )
	{
		printf( "Disaster - no MP lib loaded\n" );
		SInt16 hit;
		StandardAlert( kAlertStopAlert, "\pNo MP Library", "\pOh dear.", NULL, &hit );
		return;
	}
	dprintf(( "%d processors\n", (int)MPProcessors() ));
	
	// critical region
	err = MPCreateCriticalRegion( &m_criticalRegion );
	if ( noErr != err )
	{
		printf( "MPCreateCriticalRegion returned %d\n", (int)err );
		return;
	}
	
	// event group for notification main->thread
	err = MPCreateEvent( &m_eventGroup );
	if ( noErr != err )
	{
		printf( "MPCreateEvent returned %d\n", (int)err );
		return;
	}
	
	// semaphore for notification of task shutdown
	err = MPCreateSemaphore( 1, 0, &m_semaphore );
	if ( noErr != err )
	{
		printf( "MPCreateSemaphore returned %d\n", (int)err );
		return;
	}
	
	err = 	MPCreateTask(		  RunThread,
								  this,
								  0,
								  0,
								  0,
								  0,
								  0,
								  &m_runThreadId );
	if ( noErr != err )
	{
		printf( "MPCreateTask returned %d\n", (int)err );
		return;
	}
	dprintf(( "created task %p\n", &m_runThreadId ));
	
	SetStatus( "Welcome!" );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void	njinterface::CreatePanel( HIViewRef parentPane, WindowRef parentWindow )
{
	OSStatus err;
	
	m_ownWindow = ( parentWindow == NULL );
	if ( !m_ownWindow )
	{
		m_window = parentWindow;
		WindowAttributes attributes;
		GetWindowAttributes( m_window, &attributes );
		m_compositing = !!( attributes & kWindowCompositingAttribute );
	}
	else
	{
		m_compositing = true;
	}
	
	// get nib from bundle
	CFBundleRef bundle;
    bundle = CFBundleGetBundleWithIdentifier( CFSTR("com.ExSl.plug.ninjamplugin") );
    CFRetain( bundle );
	IBNibRef nibRef;
    CreateNibReferenceWithCFBundle( bundle, CFSTR("windows"), &nibRef );

	// make window
	WindowRef tempWindow;
	err = CreateWindowFromNib( nibRef, m_compositing ? CFSTR("mainC") : CFSTR("main"), &tempWindow );
	if ( m_ownWindow )
		m_window = tempWindow;

	// get pane from window and reparent
	ControlRef pane;
    ControlID paneID = { 'Pane', 1337 };
    GetControlByID( tempWindow, &paneID, &pane );
    Rect nibFrame;
    GetWindowBounds( tempWindow, kWindowContentRgn, &nibFrame );
	HIViewAddSubview( parentPane, pane );
    SizeControl( parentPane, nibFrame.right-nibFrame.left, nibFrame.bottom-nibFrame.top );
	if ( !m_ownWindow )
		ReleaseWindow( tempWindow );
	
	EventTypeSpec windowEventList[] = {
	{ kEventClassControl, kEventCommandProcess },
	};
	InstallWindowEventHandler( m_window, NewEventHandlerUPP( WindowEventHandler ), GetEventTypeCount( windowEventList ), windowEventList, this, NULL );

	// metronome channels
	ControlRef mech;
	ControlID mechID = { 'mech', 0 };
	GetControlByID( m_window, &mechID, &mech );
	MenuRef mechMenu = GetControlPopupMenuHandle( mech );
	DeleteMenuItems( mechMenu, 1, 1 );
	for ( int i=0; i<m_numOuts; ++i )
	{
		char buff[16];
		sprintf( buff, "%d", i+1 );
		CFStringRef str = CFStringCreateWithCString( NULL, buff, strlen(buff) );		
		MenuItemIndex dummy;
		AppendMenuItemTextWithCFString( mechMenu, str, 0, 'MECM' + i, &dummy );
		CFRelease( str );
	}
	SetControl32BitMaximum( mech, m_numOuts );
	
	int pos = 60;
	
	// make input channels
	for ( int ch=0; m_pClient->EnumLocalChannels( ch ) >= 0; ++ch )
	{
		WindowRef window;
		err = CreateWindowFromNib( nibRef, m_compositing ? CFSTR("inchannelC") : CFSTR("inchannel"), &window );
		ControlRef control;
		ControlID controlID = { 'inch', 0 };
		GetControlByID( window, &controlID, &control );
		HIViewAddSubview( pane, control );
		DisposeWindow( window );
		Rect controlFrame;
		GetControlBounds( control, &controlFrame );
		OffsetRect( &controlFrame, 0, pos );
		pos += ( controlFrame.bottom - controlFrame.top );
		SetControlBounds( control, &controlFrame );
		for ( HIViewRef sub = HIViewGetFirstSubview( control ); sub != NULL; sub = HIViewGetNextView( sub ) )
		{
			HIViewID id;
			HIViewGetID( sub, &id );
			id.id = ch;
			HIViewSetID( sub, id );
		}
		// set name
		ControlRef inna;
		ControlID innaID = { 'inna', ch };
		GetControlByID( m_window, &innaID, &inna );
		bool tx;
		char* pName = m_pClient->GetLocalChannelInfo( ch, NULL, NULL, &tx );
		SetControlData( inna, kControlEntireControl, kControlEditTextTextTag, strlen(pName), pName );
		// install callback
		SetControlReference( inna, SInt32(this) );
		ControlKeyFilterUPP proc = KeyFilterCallback;
		SetControlData( inna, 0, kControlEditTextKeyFilterTag, sizeof(proc), &proc );
		// set state
		float vol, pan;
		bool mute, solo;
		m_pClient->GetLocalChannelMonitoring( ch, &vol, &pan, &mute, &solo );
		ControlRef intx, invo, inpa, inmu, inso;
		ControlID intxID = { 'intx', ch };
		ControlID invoID = { 'invo', ch };
		ControlID inpaID = { 'inpa', ch };
		ControlID inmuID = { 'inmu', ch };
		ControlID insoID = { 'inso', ch };
		GetControlByID( m_window, &intxID, &intx );
		GetControlByID( m_window, &invoID, &invo );
		GetControlByID( m_window, &inpaID, &inpa );
		GetControlByID( m_window, &inmuID, &inmu );
		GetControlByID( m_window, &insoID, &inso );
		SetControlValue( intx, tx );
		SetControlValue( invo, pow( vol, 1.0/3 ) * GetControlMaximum( invo ) );
		SetControlValue( inpa, pan * GetControlMaximum( inpa ) );
		SetControlValue( inmu, mute );
		SetControlValue( inso, solo );
	}

	// put output channel pane inside a scroll view
	ControlRef chan;
    ControlID chanID = { 'chan', 0 };
    GetControlByID( m_window, &chanID, &chan );
	HIViewRef scroll;
	HIScrollViewCreate( kHIScrollViewOptionsVertScroll, &scroll );
	HIViewSetVisible( scroll, true );
	HIRect srect;
	srect.origin.x = 0;
	srect.origin.y = 250;
	srect.size.width = 800;
	srect.size.height = 250;
	HIViewSetFrame( scroll, &srect );
	HIViewAddSubview( pane, scroll );
	HIViewAddSubview( scroll, chan );
	
	// install send chat callback
    ControlID chinID = { 'chin', 0 };
	ControlRef chin;
    GetControlByID( m_window, &chinID, &chin );
	SetControlReference( chin, SInt32(this) );
	ControlKeyFilterUPP proc = KeyFilterCallback;
	SetControlData( chin, 0, kControlEditTextKeyFilterTag, sizeof(proc), &proc );

	ControlRef metv, metp, metm;
	ControlID metvID = { 'metv', 0 };
	ControlID metpID = { 'metp', 0 };
	ControlID metmID = { 'metm', 0 };
	GetControlByID( m_window, &metvID, &metv );
	GetControlByID( m_window, &metpID, &metp );
	GetControlByID( m_window, &metmID, &metm );
	SetControlValue( metv, pow( m_pClient->config_metronome, 1.0/3 ) * GetControlMaximum( metv ) );
	SetControlValue( metp, m_pClient->config_metronome_pan * GetControlMaximum( metp ) );
	SetControlValue( metm, m_pClient->config_metronome_mute );
	SetControl32BitValue( mech, m_pClient->config_metronome_channel + 1 );

    DisposeNibReference( nibRef );
	CFRelease( bundle );
	
	// populate server list
	CFStringRef appName = CFStringCreateWithCString( NULL, PREFS_NAME, kCFStringEncodingASCII );	// magic bug workaround
	CFPropertyListRef value = CFPreferencesCopyAppValue( CFSTR("servers"), appName );
	if ( !value )
	{
		// populate with defaults
		CFStringRef strs[5];
		strs[0] = CFSTR("exampleserver.com;username;password");
		strs[1] = CFSTR("test.ninjam.com:2049;anonymous:anonymous;");
		strs[2] = CFSTR("test.ninjam.com:2050;anonymous:anonymous;");
		strs[3] = CFSTR("test.ninjam.com:2051;anonymous:anonymous;");
		strs[4] = CFSTR("test.ninjam.com:2052;anonymous:anonymous;");
		CFArrayRef array = CFArrayCreate( NULL, (const void**)strs, 5, &kCFTypeArrayCallBacks );
		CFPreferencesSetAppValue( CFSTR("servers"), array, appName );
		CFRelease( array );
		value = CFPreferencesCopyAppValue( CFSTR("servers"), appName );
	}
	if ( value )
	{
		ControlID servID = { 'serv', 0 };
		ControlRef serv;
		GetControlByID( m_window, &servID, &serv );
		CFArrayRef array = (CFArrayRef)value;
		int num = CFArrayGetCount( array );
		for ( int i=0; i<num; ++i )
		{
			CFStringRef value = (CFStringRef)CFArrayGetValueAtIndex( array, i );
			if ( i == 0 )
			{
				char buff[256];
				CFStringGetCString( value, buff, sizeof buff, kCFStringEncodingASCII );
				SetControlData( serv, kControlEntireControl, kControlEditTextTextTag, strlen(buff), buff );
			}
			HIComboBoxAppendTextItem( serv, value, NULL );
		}
		CFRelease( value );
	}
	CFRelease( appName );

	ControlID connID = { 'conn', 0 };
	ControlRef conn;
	GetControlByID( m_window, &connID, &conn );
	if ( m_pClient->GetStatus() == NJClient::NJC_STATUS_OK )
		SetControlTitleWithCFString( conn, CFSTR("Disconnect") );
	else
		SetControlTitleWithCFString( conn, CFSTR("Connect") );
					
	m_pClient->HasUserInfoChanged();
	MakeOutputChannels();

	if ( m_ownWindow )
		ShowWindow( m_window );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void			njinterface::WindowBeingDeleted()
{
	MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
	if ( m_ownWindow )
		ReleaseWindow( m_window );
	m_window = 0;
	MPExitCriticalRegion( m_criticalRegion );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void			njinterface::MakeOutputChannels()
{
	// make output channels
	struct OutData
	{
		std::string	m_name;
		float		m_vol, m_pan;
		bool		m_sub, m_mute, m_solo;
		int			m_ch;
	};
	MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
	int numu = m_pClient->GetNumUsers();
	OutData* pUsers = new OutData[ numu ];
	int numch = 0;
	for ( int i=0; i<numu; ++i )
	{
		pUsers[i].m_name = m_pClient->GetUserState( i, &pUsers[i].m_vol, &pUsers[i].m_pan, &pUsers[i].m_mute );
		for ( int ch = 0; m_pClient->EnumUserChannels( i, ch ) >= 0; ++ch )
			numch += 1;
	}
	OutData* pChannels = new OutData[ numch ];
	int index = 0;
	for ( int i=0; i<numu; ++i )
	{
		for ( int ch = 0; m_pClient->EnumUserChannels( i, ch ) >= 0; ++ch )
		{
			OutData& d = pChannels[ index++ ];
			d.m_name = m_pClient->GetUserChannelState( i, ch, &d.m_sub, &d.m_vol, &d.m_pan, &d.m_mute, &d.m_solo, &d.m_ch );
		}
	}
	MPExitCriticalRegion( m_criticalRegion );

	CFBundleRef bundle;
    bundle = CFBundleGetBundleWithIdentifier( CFSTR("com.ExSl.plug.ninjamplugin") );
    CFRetain( bundle );
	IBNibRef nibRef;
    CreateNibReferenceWithCFBundle( bundle, CFSTR("windows"), &nibRef );

	ControlRef pane;
    ControlID paneID = { 'Pane', 1337 };
    GetControlByID( m_window, &paneID, &pane );
	ControlRef chan;
    ControlID chanID = { 'chan', 0 };
    GetControlByID( m_window, &chanID, &chan );

	HIViewSetVisible( chan, false );
	
	// remove contents
	for ( HIViewRef sub = HIViewGetFirstSubview( chan ); sub != NULL; sub = HIViewGetFirstSubview( chan ) )
		DisposeControl( sub );
	
	int pos = 0;
	index = 0;
	for ( int i=0; i<numu; ++i )
	{
		const OutData& u = pUsers[i];
		WindowRef window;
		CreateWindowFromNib( nibRef, m_compositing ? CFSTR("outuserC") : CFSTR("outuser"), &window );
		ControlRef control;
		ControlID controlID = { 'ouus', 0 };
		GetControlByID( window, &controlID, &control );
		HIViewAddSubview( chan, control );
		DisposeWindow( window );
		Rect controlFrame;
		GetControlBounds( control, &controlFrame );
		OffsetRect( &controlFrame, 0, pos );
		pos += ( controlFrame.bottom - controlFrame.top );
		SetControlBounds( control, &controlFrame );
		for ( HIViewRef sub = HIViewGetFirstSubview( control ); sub != NULL; sub = HIViewGetNextView( sub ) )
		{
			HIViewID id;
			HIViewGetID( sub, &id );
			id.id = i;
			HIViewSetID( sub, id );
		}
		// set name
		ControlRef uuna;
		ControlID uunaID = { 'uuna', i };
		GetControlByID( m_window, &uunaID, &uuna );
		SetControlData( uuna, kControlEntireControl, kControlEditTextTextTag, u.m_name.size(), u.m_name.c_str() );
		for ( int ch = 0; m_pClient->EnumUserChannels( i, ch ) >= 0; ++ch )
		{
			const OutData& d = pChannels[ index++ ];

			CreateWindowFromNib( nibRef, m_compositing ? CFSTR("outchannelC") : CFSTR("outchannel"), &window );
			ControlRef control;
			ControlID controlID = { 'ouch', 0 };
			GetControlByID( window, &controlID, &control );
			HIViewAddSubview( chan, control );
			DisposeWindow( window );
			Rect controlFrame;
			GetControlBounds( control, &controlFrame );
			OffsetRect( &controlFrame, 0, pos );
			pos += ( controlFrame.bottom - controlFrame.top );
			SetControlBounds( control, &controlFrame );
			int cid = (i << 16) + ch;
			for ( HIViewRef sub = HIViewGetFirstSubview( control ); sub != NULL; sub = HIViewGetNextView( sub ) )
			{
				HIViewID id;
				HIViewGetID( sub, &id );
				id.id = cid;
				HIViewSetID( sub, id );
			}
			// set name
			ControlRef ouna;
			ControlID ounaID = { 'ouna', cid };
			GetControlByID( m_window, &ounaID, &ouna );
			SetControlData( ouna, kControlEntireControl, kControlEditTextTextTag, d.m_name.size(), d.m_name.c_str() );
			// channel menu
			ControlRef outc;
			ControlID outcID = { 'outc', cid };
			GetControlByID( m_window, &outcID, &outc );
			MenuRef outcMenu = GetControlPopupMenuHandle( outc );
			DeleteMenuItems( outcMenu, 1, 1 );
			for ( int i=0; i<m_numOuts; ++i )
			{
				char buff[16];
				sprintf( buff, "%d", i+1 );
				CFStringRef str = CFStringCreateWithCString( NULL, buff, strlen(buff) );		
				MenuItemIndex dummy;
				AppendMenuItemTextWithCFString( outcMenu, str, 0, 'OUTC' + i, &dummy );
				CFRelease( str );
			}
			SetControl32BitMaximum( outc, m_numOuts );
			SetControl32BitValue( outc, d.m_ch + 1 );
			// set state
			ControlRef ourx, ouvo, oupa, oumu, ouso;
			ControlID ourxID = { 'ourx', cid };
			ControlID ouvoID = { 'ouvo', cid };
			ControlID oupaID = { 'oupa', cid };
			ControlID oumuID = { 'oumu', cid };
			ControlID ousoID = { 'ouso', cid };
			GetControlByID( m_window, &ourxID, &ourx );
			GetControlByID( m_window, &ouvoID, &ouvo );
			GetControlByID( m_window, &oupaID, &oupa );
			GetControlByID( m_window, &oumuID, &oumu );
			GetControlByID( m_window, &ousoID, &ouso );
			SetControlValue( ourx, d.m_sub );
			SetControlValue( ouvo, pow( d.m_vol, 1.0/3 ) * GetControlMaximum( ouvo ) );
			SetControlValue( oupa, d.m_pan * GetControlMaximum( oupa ) );
			SetControlValue( oumu, d.m_mute );
			SetControlValue( ouso, d.m_solo );
		}
	}
	delete [] pChannels;
	delete [] pUsers;
	
	DisposeNibReference( nibRef );
	CFRelease( bundle );

	HIRect srect;
	srect.origin.x = 0;
	srect.origin.y = 0;
	srect.size.width = 780;
	srect.size.height = pos;
	HIViewSetFrame( chan, &srect );
	HIViewSetVisible( chan, true );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

njinterface::~njinterface()
{
	dprintf(( "Destroying task %p\n", &m_runThreadId ));
	
	// send event
	MPSetEvent( m_eventGroup, kEventTerminate );
	// wait for notification
	if ( kMPTimeoutErr == MPWaitOnSemaphore( m_semaphore, kDurationMillisecond * 1000 * 10 ) )
	{
		printf( "DestroyRenderTask timeout on shutdown\n" );
		MPTerminateTask( m_runThreadId, 0 );
		m_runThreadId = 0;
	}
	
	// clean up
	MPDeleteEvent( m_eventGroup );
	MPDeleteSemaphore( m_semaphore );
	MPDeleteCriticalRegion( m_criticalRegion );
	
	dprintf(( "    done\n" ));

	// disconnect
	m_pClient->Disconnect();
	
	delete m_pClient;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSStatus			njinterface::RunThread( void *pParam )
{
	return ((njinterface *)pParam)->RunThreadMain();
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

OSStatus			njinterface::RunThreadMain()
{
	uint tick = 0;
	
	int lastStatus = 999, lastBpi, lastPos, lastLen;
	float lastBpm = 0.0f;
	std::string lastUser, lastHost, connectionInfo;
	
	for (;;)
	{
/*
		if (!( tick % 100 ))
		{
			MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
			printf( "status: %d %d %s\n", m_pClient->GetStatus(), m_pClient->IsAudioRunning(), m_pClient->GetErrorStr() );
			printf( "local peaks: %f %f\n", m_pClient->GetLocalChannelPeak( 0 ), m_pClient->GetLocalChannelPeak( 1 ) );
			int num = m_pClient->GetNumUsers();
			for ( int i=0; i<num; ++i )
			{
				float vol, pan;
				bool mute;
				char* pName = m_pClient->GetUserState( i, &vol, &pan, &mute );
				printf( "user %d: %s: vol %f pan %f mute %d\n", i, pName, vol, pan, mute );
				for ( int ch = 0; m_pClient->EnumUserChannels( i, ch ) >= 0; ++ch )
				{
					m_pClient->SetUserChannelState( i, ch, false, 0, false, 0, false, 0, false, 0, false, 0, true, std::min( i*2, 8 ) );
					bool sub, solo;
					int outch;
					pName = m_pClient->GetUserChannelState( i, ch, &sub, &vol, &pan, &mute, &solo, &outch );
					printf( "chan %d: %s: sub %d vol %f pan %f mute %d solo %d: outch %d: peak %f\n", ch, pName, sub, vol, pan, mute, solo, outch, m_pClient->GetUserChannelPeak( i, 0 ) );
				}
			}
			MPExitCriticalRegion( m_criticalRegion );
		}
*/
		if (!( tick & 3 ))
		{
			MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
			int num = 0;
			int numu = m_pClient->GetNumUsers();
			for ( int i=0; i<numu; ++i )
			{
				for ( int ch = 0; m_pClient->EnumUserChannels( i, ch ) >= 0; ++ch )
					num++;
			}
			int* ids = (int*)alloca( num * sizeof(int) );
			float* levels = (float*)alloca( num * sizeof(float) );
			int index = 0;
			for ( int i=0; i<numu; ++i )
			{
				for ( int ch = 0; m_pClient->EnumUserChannels( i, ch ) >= 0; ++ch )
				{
					ids[index] = (i<<16)+ch;
					levels[index++] = m_pClient->GetUserChannelPeak( i, ch );
				}
			}
			MPExitCriticalRegion( m_criticalRegion );
			for ( int i=0; i<index; ++i )
			{
				ControlRef oudb;
				ControlID oudbID = { 'oudb', ids[i] };
				GetControlByID( m_window, &oudbID, &oudb );
				double db = VAL2DB( levels[i] );
				char buff[32] = "-Inf dB";
				if ( db > -120.0 )
					sprintf( buff, "%+.1f dB", db );
				SetControlData( oudb, kControlEntireControl, kControlEditTextTextTag, strlen(buff), buff );
			}
			for ( int ch = 0; m_pClient->EnumLocalChannels( ch ) >= 0; ++ch )
			{
				ControlRef indb;
				ControlID indbID = { 'indb', ch };
				GetControlByID( m_window, &indbID, &indb );
				double db = m_pClient->GetLocalChannelPeak( ch );
				db = VAL2DB( db );
				char buff[32] = "-Inf dB";
				if ( db > -120.0 )
					sprintf( buff, "%+.1f dB", db );
				SetControlData( indb, kControlEntireControl, kControlEditTextTextTag, strlen(buff), buff );
			}
		}
		
		// update status
		MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
		int status = m_pClient->GetStatus();
		int bpi = m_pClient->GetBPI();
		float bpm = m_pClient->GetActualBPM();
		int pos, len;
		m_pClient->GetPosition( &pos, &len );
		pos = 1 + (int)((pos * bpi)/len);
		char* user = m_pClient->GetUserName();
		char* host = m_pClient->GetHostName();
		if ( lastUser != user || lastHost != host )
		{
			lastUser = user;
			lastHost = host;
			connectionInfo = lastUser + " @ " + lastHost;
		}
		MPExitCriticalRegion( m_criticalRegion );
		if ( lastStatus != status )
		{
			lastStatus = status;
			if ( status == NJClient::NJC_STATUS_CANTCONNECT )
				connectionInfo = "Can't connect";
			else if ( status == NJClient::NJC_STATUS_DISCONNECTED )
				connectionInfo = "Disconnected";
			else if ( status == NJClient::NJC_STATUS_INVALIDAUTH )
				connectionInfo = "Authorisation failed";
		}
		if ( lastBpi != bpi || lastBpm != bpm )
		{
			lastBpi = bpi;
			lastBpm = bpm;
			ControlID bpimID = { 'bpim', 0 };
			ControlRef bpim;
			GetControlByID( m_window, &bpimID, &bpim );
			char buff[100];
			sprintf( buff, "%.2f bpm", bpm );
			SetControlData( bpim, kControlEntireControl, kControlEditTextTextTag, strlen(buff), buff );
		}
		if ( lastPos != pos || lastLen != len )
		{
			lastPos = pos;
			lastLen = len;
			ControlID poslID = { 'posl', 0 };
			ControlRef posl;
			GetControlByID( m_window, &poslID, &posl );
			char buff[100];
			sprintf( buff, "%2d/%2d", pos, bpi );
			SetControlData( posl, kControlEntireControl, kControlEditTextTextTag, strlen(buff), buff );
		}
		if ( !connectionInfo.empty() )
			SetStatus( connectionInfo.c_str() );
		
		if ( m_pClient->HasUserInfoChanged() )
			MakeOutputChannels();

		// let the client run
		std::string goodServer;
		bool setConn = false;
		MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
		while ( !m_pClient->Run() ); 
		if ( !m_connectionAttempt.empty() )
		{
			int status = m_pClient->GetStatus();
			if ( status <= 0 )
			{
				setConn = true;
				if ( status == NJClient::NJC_STATUS_OK )
					goodServer = m_connectionAttempt;
				m_connectionAttempt.clear();
			}
		}
		if ( setConn )
		{
			ControlID connID = { 'conn', 0 };
			ControlRef conn;
			GetControlByID( m_window, &connID, &conn );
			if ( !goodServer.empty() )
			{
				SetControlTitleWithCFString( conn, CFSTR("Disconnect") );
				// add good servers to list
				CFStringRef value = CFStringCreateWithCString( NULL, goodServer.c_str(), kCFStringEncodingASCII );
				ControlID servID = { 'serv', 0 };
				ControlRef serv;
				GetControlByID( m_window, &servID, &serv );
				int num = HIComboBoxGetItemCount( serv );
				CFStringRef* strs = new CFStringRef[num+1];
				bool exists = false;
				for ( int i=0; i<num; ++i )
				{
					HIComboBoxCopyTextItemAtIndex( serv, i, &strs[i] );
					if ( kCFCompareEqualTo == CFStringCompare( strs[i], value, kCFCompareCaseInsensitive ) )
						exists = true;
				}
				if ( !exists )
				{
					HIComboBoxInsertTextItemAtIndex( serv, 0, value );
					for ( int i=num-1; i>=0; --i )
						strs[i+1] = strs[i];
					HIComboBoxCopyTextItemAtIndex( serv, 0, &strs[0] );
					num += 1;
				}
				CFRelease( value );
				// copy out server list to prefs
				CFStringRef appName = CFStringCreateWithCString( NULL, PREFS_NAME, kCFStringEncodingASCII );	// magic bug workaround
				CFArrayRef array = CFArrayCreate( NULL, (const void**)strs, num, &kCFTypeArrayCallBacks );
				CFPreferencesSetAppValue( CFSTR("servers"), array, appName );
				CFPreferencesAppSynchronize( appName );
				for ( int i=0; i<num; ++i )
					CFRelease( strs[i] );
				delete [] strs;
				CFRelease( array );
				CFRelease( appName );
			}
			else
				SetControlTitleWithCFString( conn, CFSTR("Connect") );
		}
		MPExitCriticalRegion( m_criticalRegion );
			
		// wait for an event, or timeout
		MPEventFlags flags = 0;
		MPWaitForEvent( m_eventGroup, &flags, kDurationMillisecond * 35 );
		if ( flags & kEventTerminate )
			break;
		
		tick++;
	}
	
	// before terminating the task, let the main thread know we're done
	MPSignalSemaphore( m_semaphore );
	
	return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

int					njinterface::LicenseCallback( int user32, char *licensetext )
{
	njinterface* pThis = (njinterface*)user32;

	// !!!!!!! temporary workaround, see below
	return true;
	
// TODO - if anyone can figure out why the below crashes, I'll be very greatful.
#if 0
	CFBundleRef bundle;
    bundle = CFBundleGetBundleWithIdentifier( CFSTR("com.ExSl.plug.ninjamplugin") );
    CFRetain( bundle );
    
	IBNibRef nibRef;
    CreateNibReferenceWithCFBundle( bundle, CFSTR("windows"), &nibRef );
	
	WindowRef window;
	CreateWindowFromNib( nibRef, CFSTR("license"), &window );
	
	EventTypeSpec windowEventList[] = {
	{ kEventClassControl, kEventCommandProcess },
	};
	
    DisposeNibReference( nibRef );
	CFRelease( bundle );

	ControlID textID = { 'text', 0 };
	ControlRef text;
    GetControlByID( window, &textID, &text );
	TXNObject txn = HITextViewGetTXNObject(text);
	TXNSetData( txn, kTXNTextData, licensetext, strlen(licensetext), kTXNEndOffset, kTXNEndOffset );

	pThis->m_licenceResult = 0;
	InstallWindowEventHandler( window, NewEventHandlerUPP( LicenseWindowEventHandler ), GetEventTypeCount( windowEventList ), windowEventList, pThis, NULL );
	
	ShowWindow( window );
	
	MPExitCriticalRegion( pThis->m_criticalRegion );
	while ( *(volatile int*)&pThis->m_licenceResult == 0 )
	{
		struct timespec ts={0,10000*1000};
		nanosleep(&ts,NULL);
	}
	MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );

	return ( 1 == pThis->m_licenceResult );
#endif
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void				njinterface::ChatMessageCallback(int user32, NJClient *inst, char **parms, int nparms)
{
	njinterface* pThis = (njinterface*)user32;

    ControlID chatID = { 'chat', 0 };
	ControlRef chat;
    GetControlByID( pThis->m_window, &chatID, &chat );
	
	TXNObject txn = HITextViewGetTXNObject(chat);

	for ( int i=0; i<nparms; ++i )
	{
		int len = strlen( parms[i] );
		if ( len > 0 )
			TXNSetData( txn, kTXNTextData, parms[i], len, kTXNEndOffset, kTXNEndOffset );
		TXNSetData( txn, kTXNTextData, " ", 1, kTXNEndOffset, kTXNEndOffset );
	}
	TXNSetData( txn, kTXNTextData, "\n", 1, kTXNEndOffset, kTXNEndOffset );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void				njinterface::AudioProc( const float** pIns, float** pOuts, int numIns, int numOuts, UInt32 inFramesToProcess, int sampleRate )
{
	// cache sample rate
	m_sampleRate = sampleRate;
	
	m_pClient->AudioProc( (float**)pIns, numIns, pOuts, numOuts, inFramesToProcess, sampleRate );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pascal OSStatus		njinterface::WindowEventHandler( EventHandlerCallRef  inHandler, 
												EventRef             inEvent, 
												void*                userData )
{
	njinterface* pThis = (njinterface*)userData;

	ControlRef control;
    WindowRef wind;
    
    // get control hit by event
    GetEventParameter( inEvent, kEventParamDirectObject, typeControlRef, NULL, sizeof(ControlRef), NULL, &control );
	ControlID id;
	GetControlID( control, &id );
    wind = GetControlOwner( control );
	
	switch( id.signature )
	{
		case 'serv':
			printf( "serv\n" );
			break;
		case 'conn':
			pThis->Connect();
			break;
		case 'metm':
			// mute metronome
		{
			int v = GetControlValue( control );
			pThis->m_pClient->config_metronome_mute = ( v != 0 );
		}
			break;
		case 'metv':
			// metronome volume
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			v = v*v*v;
			pThis->m_pClient->config_metronome = v;
		}
			break;
		case 'metp':
			// metronome pan
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			pThis->m_pClient->config_metronome_pan = v;
		}
			break;
		case 'invo':
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			v = v*v*v;
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetLocalChannelMonitoring( id.id, true, v, false, 0, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'inpa':
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetLocalChannelMonitoring( id.id, false, 0, true, v, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'inmu':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetLocalChannelMonitoring( id.id, false, 0, false, 0, true, v, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'inso':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetLocalChannelMonitoring( id.id, false, 0, false, 0, false, 0, true, v );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'intx':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			std::string name = pThis->m_pClient->GetLocalChannelInfo( id.id, 0, 0, 0 );
			pThis->m_pClient->SetLocalChannelInfo( id.id, (char*)name.c_str(), false, 0, false, 0, true, v );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'uuvo':
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			v = v*v*v;
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserState( id.id, true, v, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'uupa':
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserState( id.id, false, 0, true, v, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'uumu':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserState( id.id, false, 0, false, 0, true, v );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'ouvo':
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			v = v*v*v;
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserChannelState( id.id>>16, id.id&0xffff, false, 0, true, v, false, 0, false, 0, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'oupa':
		{
			float v = GetControlValue( control )/(float)GetControlMaximum( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserChannelState( id.id>>16, id.id&0xffff, false, 0, false, 0, true, v, false, 0, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'oumu':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserChannelState( id.id>>16, id.id&0xffff, false, 0, false, 0, false, 0, true, v, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'ouso':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserChannelState( id.id>>16, id.id&0xffff, false, 0, false, 0, false, 0, false, 0, true, v, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'ourx':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserChannelState( id.id>>16, id.id&0xffff, true, v, false, 0, false, 0, false, 0, false, 0, false, 0 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'outc':
		{
			int v = GetControlValue( control );
			MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
			pThis->m_pClient->SetUserChannelState( id.id>>16, id.id&0xffff, false, 0, false, 0, false, 0, false, 0, false, 0, true, v-1 );
			MPExitCriticalRegion( pThis->m_criticalRegion );
		}
			break;
		case 'mech':
		{
			int v = GetControlValue( control );
			pThis->m_pClient->config_metronome_channel = v - 1;
		}
			break;
	}
	
	return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pascal OSStatus		njinterface::LicenseWindowEventHandler( EventHandlerCallRef  inHandler, 
															EventRef             inEvent, 
															void*				userData )
{
	njinterface* pThis = (njinterface*)userData;
	
	ControlRef control;
    UInt32 cmd;
    
    // get control hit by event
    GetEventParameter( inEvent, kEventParamDirectObject, typeControlRef, NULL, sizeof(ControlRef), NULL, &control );
    // get the command for that control
    GetControlCommandID( control, &cmd );
	
	switch( cmd )
	{
		case 'agre':
			pThis->m_licenceResult = 1;
			DisposeWindow( GetControlOwner( control ) );
			break;
		case 'decl':
			pThis->m_licenceResult = 2;
			DisposeWindow( GetControlOwner( control ) );
			break;
	}
	
	return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

pascal ControlKeyFilterResult	njinterface::KeyFilterCallback( ControlRef control, 
														   SInt16 *keyCode, SInt16 *charCode,
														   EventModifiers *modifiers )
{
	SInt16 c = *charCode;
	if ( c == '\r' || c == 3 ) 
	{
		njinterface *pThis = (njinterface*)GetControlReference( control );
		ControlID id;
		GetControlID( control, &id );
		switch ( id.signature )
		{
			case 'inna':
			{
				Size size;
				GetControlDataSize( control, kControlEntireControl, kControlEditTextTextTag, &size );
				char* buffer = new char[ size+1 ];
				GetControlData( control, kControlEntireControl, kControlEditTextTextTag, size, buffer, &size );
				buffer[size] = 0;
				MPEnterCriticalRegion( pThis->m_criticalRegion, kDurationForever );
				pThis->m_pClient->SetLocalChannelInfo( id.id, buffer, false, 0, false, 0, false, 0 );
				MPExitCriticalRegion( pThis->m_criticalRegion );
				delete [] buffer;
				ControlEditTextSelectionRec sel = { 0, 32767 };
				SetControlData( control, 0, kControlEditTextSelectionTag, sizeof(sel), &sel );
			}
				return kControlKeyFilterBlockKey;
			case 'chin':
				pThis->SendChat();
				return kControlKeyFilterBlockKey;
		}
	}
	return kControlKeyFilterPassKey;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void		njinterface::SendChat()
{
	// get text
    ControlID chinID = { 'chin', 0 };
	ControlRef chin;
    GetControlByID( m_window, &chinID, &chin );
	Size size;
	GetControlDataSize( chin, kControlEntireControl, kControlEditTextTextTag, &size );
	char* buffer = new char[ size+1 ];
	GetControlData( chin, kControlEntireControl, kControlEditTextTextTag, size, buffer, &size );
	buffer[size] = 0;
	
	// send text
	if ( size > 0 )
	{
		MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
		if ( buffer[0] == '/' )
			m_pClient->ChatMessage_Send( "ADMIN", buffer+1 );
		else
			m_pClient->ChatMessage_Send( "MSG", buffer );
		MPExitCriticalRegion( m_criticalRegion );
	}
	
	delete [] buffer;
	
	// clear control
	SetControlData( chin, kControlEntireControl, kControlEditTextTextTag, 0, "" );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void		njinterface::Connect()
{
    ControlID connID = { 'conn', 0 };
	ControlRef conn;
    GetControlByID( m_window, &connID, &conn );

	MPEnterCriticalRegion( m_criticalRegion, kDurationForever );
	int status = m_pClient->GetStatus();
	if ( status != NJClient::NJC_STATUS_OK )
	{
		// connect
		ControlID servID = { 'serv', 0 };
		ControlRef serv;
		GetControlByID( m_window, &servID, &serv );
		Size size;
		GetControlDataSize( serv, kControlEntireControl, kControlEditTextTextTag, &size );
		char* pBuffer = new char[ size+1 ];
		GetControlData( serv, kControlEntireControl, kControlEditTextTextTag, size, pBuffer, &size );
		pBuffer[size] = 0;
		printf( "connect to: %s\n", pBuffer );
		m_connectionAttempt = pBuffer;
		char* pPass = pBuffer;
		char* pHost = strsep( &pPass, ";" );
		char* pUser = strsep( &pPass, ";" );
		if ( pHost && pUser && pPass )
		{
			m_pClient->Connect( pHost, pUser, pPass );
			SetControlTitleWithCFString( conn, CFSTR("Connecting...") );
		}
		else
		{
			m_connectionAttempt.clear();
		}
		delete [] pBuffer;
	}
	else
	{
		// disconnect
		m_pClient->Disconnect();
		SetControlTitleWithCFString( conn, CFSTR("Connect") );
	}
	MPExitCriticalRegion( m_criticalRegion );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

void		njinterface::SetStatus( const char* pString )
{
	ControlID statID = { 'stat', 0 };
	ControlRef stat;
	GetControlByID( m_window, &statID, &stat );
	SetControlData( stat, kControlEntireControl, kControlEditTextTextTag, strlen(pString), pString );
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
